import typing

from libbs.artifacts import Comment

from .prompt import Prompt

if typing.TYPE_CHECKING:
    from dailalib.api import AIAPI


def rename_function(result, function, ai_api: "AIAPI"):
    new_name = list(result.values())[0]
    function.name = new_name
    ai_api._dec_interface.functions[function.addr] = function


def rename_variables(result, function, ai_api: "AIAPI"):
    ai_api._dec_interface.rename_local_variables_by_names(function, result)


def comment_function(result, function, ai_api: "AIAPI"):
    ai_api._dec_interface.comments[function.addr] = Comment(function.addr, comment=result, func_addr=function.addr)


PROMPTS = [
    Prompt(
        "summarize",
        f'''
        You are C expert that summarizes code. When given code, you summarize at a high level what the function does 
        and you identify if known algorithms are used in the function. As an example:
        """
        int sub_404000(int a0, char** a1) 
        {{
            int v1; // rax 
            v1 = sub_404100(a0[1]) % 2 == 0 
            return v1;
        }}
        """

        You responded with:
        This function takes two arguments and implements the is_even check on second argument

        Here is another example:
        {Prompt.DECOMP_TEXT}

        You responded with:
        ''',
        desc="Summarize the function",
        gui_result_callback=comment_function
    ),
    Prompt(
        "identify_source",
        f'''
        You are a C expert that identifies the original source given decompilation. Upon discovering the source,
        you give a link to the code. You only respond with a json. As an example:
        """
        void __fastcall __noreturn usage(int status)
        {{
            v2 = program_name;
            if ( status )
            {{
                v3 = dcgettext(0LL, "Try '%s --help' for more information.\n", 5);
                _fprintf_chk(stderr, 1LL, v3, v2);
            }}
            // ...
        }}
        """

        You responded with:
        {{"link": "https://www.gnu.org/software/coreutils/"}}

        When you don't know, you respond with {{}}.

        Here is another example:
        {Prompt.DECOMP_TEXT}

        You responded with:
        ''',
        desc="Identify the original source",
        json_response=True,
        gui_result_callback=comment_function
    ),
    Prompt(
        "rename_variables",
        f'''
        You are C expert that renames variables in code. When given code, you rename variables according to the
        meaning of the function or its use. You only respond with a json. As an example:

        int sub_404000(int a0, char** a1) 
        {{
            int v1; // rax 
            v1 = sub_404100(a0[1]) % 2 == 0 
            return v1;
        }}

        You responded with:
        {{"a0: argc", "a1: argv",  "v1": "is_even"}}

        Here is another example:
        {Prompt.DECOMP_TEXT}

        You responded with:
        ''',
        desc="Suggest variable names",
        json_response=True,
        gui_result_callback=rename_variables,
    ),
    Prompt(
        "rename_function",
        f'''
        You are C expert that renames functions. When given a function, you rename it according to the
        meaning of the function or its use. You only respond with a json. As an example:

        int sub_404000(int a0, char** a1) 
        {{
            int is_even; // rax 
            is_even = sub_404100(a0[1]) % 2 == 0 
            return is_even;
        }}

        You responded with:
        {{"sub_404000": "number_is_even"}}

        Here is another example:
        {Prompt.DECOMP_TEXT}

        You responded with:
        ''',
        desc="Suggest a function name",
        json_response=True,
        gui_result_callback=rename_function
    ),
]
PROMPTS_BY_NAME = {p.name: p for p in PROMPTS}



import json
import re
from typing import Optional, Union, Dict, Callable
import textwrap

from dailalib.api import AIAPI
from dailalib.api.openai.openai_api import OpenAIAPI

JSON_REGEX = re.compile(r"\{.*}", flags=re.DOTALL)


class Prompt:
    DECOMP_REPLACEMENT_LABEL = "<DECOMPILATION>"
    SNIPPET_REPLACEMENT_LABEL = "<SNIPPET>"
    SNIPPET_TEXT = f"\n\"\"\"{SNIPPET_REPLACEMENT_LABEL}\"\"\""
    DECOMP_TEXT = f"\n\"\"\"{DECOMP_REPLACEMENT_LABEL}\"\"\""

    def __init__(
        self,
        name: str,
        text: str,
        desc: str = None,
        pretext_response: Optional[str] = None,
        posttext_response: Optional[str] = None,
        json_response: bool = False,
        ai_api=None,
        # callback(result, function, ai_api)
        gui_result_callback: Optional[Callable] = None
    ):
        self.name = name
        self.text = textwrap.dedent(text)
        self._pretext_response = pretext_response
        self._posttext_response = posttext_response
        self._json_response = json_response
        self._gui_result_callback = gui_result_callback
        self.desc = desc or name
        self.ai_api: "OpenAIAPI" = ai_api

    def query_model(self, *args, function=None, dec_text=None, use_dec=True, **kwargs):
        if self.ai_api is None:
            raise Exception("api must be set before querying!")

        @AIAPI.requires_function
        def _query_model(ai_api=self.ai_api, function=function, dec_text=dec_text, use_dec=use_dec) -> Union[Dict, str]:
            if not ai_api:
                return {}

            ai_api.info(f"Querying {self.name} prompt with function {function}...")
            response = self._pretext_response if self._pretext_response and not self._json_response else ""
            # grab decompilation and replace it in the prompt, make sure to fix the decompilation for token max
            query_text = self.text.replace(
                self.DECOMP_REPLACEMENT_LABEL,
                OpenAIAPI.fit_decompilation_to_token_max(dec_text)
            )
            response += self.ai_api.query_model(query_text)
            default_response = {} if self._json_response else ""
            if not response:
                return default_response

            # changes response type to a dict
            if self._json_response:
                # if the response of OpenAI gets cut off, we have an incomplete JSON
                if "}" not in response:
                    response += "}"

                json_matches = JSON_REGEX.findall(response)
                if not json_matches:
                    return default_response

                json_data = json_matches[0]
                try:
                    response = json.loads(json_data)
                except Exception:
                    response = {}
            else:
                response += self._posttext_response if self._pretext_response else ""

            if ai_api.has_decompiler_gui and response:
                self._gui_result_callback(response, function, ai_api)

            ai_api.info(f"Reponse recieved...")
            return response
        return _query_model(ai_api=self.ai_api, function=function, dec_text=dec_text, use_dec=use_dec)

